use std::collections::HashMap;

const ASXVERSION: [u8; 3] = [0,1,0];

#[derive(Debug, Clone)]
pub enum VMMemory {
    String(VMMemoryString),
    Number(VMMemoryNumber),
    Boolean(VMMemoryBoolean),
    Null(VMMemoryNull),
}

#[derive(Debug, Clone)]
struct VMMemoryString {
    value: String,
    variable_id: u32,
}
#[derive(Debug, Clone)]
struct VMMemoryNumber {
    value: i64,
    variable_id: u32,
}
#[derive(Debug, Clone)]
struct VMMemoryBoolean {
    value: bool,
    variable_id: u32,
}
#[derive(Debug, Clone)]
struct VMMemoryNull {
    variable_id: u32,
}

#[derive(Debug, Clone)]
struct DecompiledFile {
    name: String,
    functions: Vec<DecompiledFunction>,
}
#[derive(Debug, Clone)]
struct DecompiledFunction {
    name: String,
    body: Vec<DecompiledOperation>,
    var_ids: HashMap<String, u32>,
    strings: HashMap<u32, String>,
    functions: HashMap<u32, u32>
}
#[derive(Debug, Clone)]
struct DecompiledOperation {
    opcode: u8,
    arg1: u8,
    arg2: i64,
    arg3: u8,
}

#[derive(Debug)]
pub struct Register {
    id: u8,
    pointer: usize
}

pub struct Machine {
    pub memory: Vec<VMMemory>,
    pub files: Vec<DecompiledFile>,
    pub stack: Vec<VMMemory>,
    pub registers: Vec<Register>,
    pub pc: usize,
}

fn read_be_num(input: &[u8]) -> usize {
    let mut result: usize = 0;
    for i in 0..input.len() {
        result <<= 8;
        result |= input[i] as usize;
    }
    return result;
}

fn read_str(input: &[u8]) -> String {
    let mut result = String::new();
    for &byte in input {
        result.push(byte as char);
    }
    return result;
}

fn read_bin(input: &[u8]) -> String {
    let mut result = String::new();
    for &byte in input {
        result.push_str(&format!("{:08b}", byte));
    }
    return result;
}

fn bin_to_num(bin: String) -> usize {
    let mut result: usize = 0;
    for (i, c) in bin.chars().rev().enumerate() {
        if c == '1' {
            result += 1 << i;
        }
    }
    return result;
}

fn bin_to_snum(bin: String) -> i64 {
    let mut result: i64 = 0;
    let mut is_negative = false;
    for (i, c) in bin.chars().rev().enumerate() {
        if c == '1' {
            if i == bin.len() - 1 {
                is_negative = true;
            } else {
                result += 1 << i;
            }
        }
    }
    if is_negative {
        result = -result;
    }
    return result;
}

fn get_file_by_name(files: &Vec<DecompiledFile>, name: &str) -> Option<DecompiledFile> {
    for file in files {
        if file.name == name {
            return Some(file.clone());
        }
    }
    None
}

fn get_function_by_name(file: &DecompiledFile, name: &str) -> Option<DecompiledFunction> {
    for function in &file.functions {
        if function.name == name {
            return Some(function.clone());
        }
    }
    None
}

fn get_register_by_id(registers: &Vec<Register>, id: u8) -> Option<&Register> {
    for reg in registers {
        if reg.id == id {
            return Some(reg);
        }
    }
    None
}

fn set_register(registers: &mut Vec<Register>, new_register: Register) {
    if new_register.id == 0 {
        return;
    }
    for (i, reg) in registers.iter().enumerate() {
        if reg.id == new_register.id {
            registers[i] = new_register;
            return;
        }
    }
}

impl Machine {
    pub fn new() -> Self {
        let mut registers = Vec::new();
        for i in 1..17 {
            registers.push(Register {
                id: i,
                pointer: 0,
            });
        }
        return Machine {
            memory: vec![
                VMMemory::Null(VMMemoryNull { variable_id: 0 })
            ],
            files: Vec::new(),
            stack: Vec::new(),
            registers: registers,
            pc: 0,
        };
    }

    pub fn load(&mut self, data: Vec<u8>) {
        if data[0..3] != *"ASX".as_bytes() {
            panic!("Invalid ASX file header");
        }
        if data[3..6] != ASXVERSION {
            panic!("Unsupported ASX version");
        }
        let filename_len = read_be_num(&data[6..10]);
        let filename = read_str(&data[10..10 + filename_len]);
        let mut offset = 10 + filename_len;
        let func_name_len = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        let func_name = read_str(&data[offset..offset + func_name_len]);
        offset += func_name_len;
        let var_id_len = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        let mut var_ids: HashMap<String, u32> = HashMap::new();
        for _ in 0..var_id_len {
            let name_len = read_be_num(&data[offset..offset + 4]);
            offset += 4;
            let name = read_str(&data[offset..offset + name_len]);
            offset += name_len;
            let var_id = read_be_num(&data[offset..offset + 3]) as u32;
            offset += 3;
            var_ids.insert(name, var_id);
        }
        let mut strings: HashMap<u32, String> = HashMap::new();
        let string_count = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        for _ in 0..string_count {
            let str_id = read_be_num(&data[offset..offset + 3]) as u32;
            offset += 3;
            let str_len = read_be_num(&data[offset..offset + 4]);
            offset += 4;
            let str_value = read_str(&data[offset..offset + str_len]);
            offset += str_len;
            strings.insert(str_id, str_value);
        }
        let mut functions: HashMap<u32, u32> = HashMap::new();
        let _function_count = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        for _ in 0.._function_count {
            let func_id = read_be_num(&data[offset..offset + 3]) as u32;
            offset += 3;
            let func_pos = read_be_num(&data[offset..offset + 4]) as u32;
            offset += 4;
            functions.insert(func_id, func_pos);
        }
        let mut body: Vec<DecompiledOperation> = Vec::new();
        let instr_len: usize = read_be_num(&data[offset..offset + 4]);
        offset += 4;
        for _ in 0..instr_len {
            let op = read_bin(&data[offset..offset + 10]);
            offset += 10;
            println!("{}", op);
            let op_code = bin_to_num(op[0..5].to_string());
            let arg1 = bin_to_num(op[5..9].to_string());
            let arg2 = bin_to_snum(op[9..73].to_string());
            let arg3 = bin_to_num(op[73..77].to_string());
            body.push(DecompiledOperation {
                opcode: op_code as u8,
                arg1: arg1 as u8,
                arg2: arg2,
                arg3: arg3 as u8,
            });
        }
        let mut file_functions: Vec<DecompiledFunction> = Vec::new();
        file_functions.push(DecompiledFunction { name: func_name, body: body, var_ids: var_ids, strings: strings, functions: functions });
        self.files.push(DecompiledFile {
            name: filename,
            functions: file_functions
        });
        println!("Decompiled: {:?}", self.files[0]);
    }

    pub fn run(&mut self) {
        let main = get_file_by_name(&self.files, "main.asl");
        if main.is_none() {
            panic!("Main file not found");
        }
        let main_func = get_function_by_name(&main.unwrap(), "main");
        if main_func.is_none() {
            panic!("Main function not found");
        }
        let main_func = main_func.unwrap();
        while self.pc < main_func.body.len() {
            let operation = &main_func.body[self.pc];
            self.pc += 1;
            match operation.opcode {
                0 => {
                    //HLT
                    return;
                },
                3 => {
                    //LDI
                    self.memory.push(VMMemory::Number(VMMemoryNumber { value: operation.arg2, variable_id: 0 }));
                    set_register(&mut self.registers, Register { id: operation.arg1, pointer: self.memory.len() - 1 });
                },
                7 => {
                    //ASS
                    let reg = get_register_by_id(&self.registers, operation.arg1);
                    if reg.is_none() {
                        panic!("Register {} not found", operation.arg1);
                    }
                    let reg = reg.unwrap();
                    if reg.pointer >= self.memory.len() || reg.pointer < 1 {
                        panic!("Register {} points to an invalid memory location", operation.arg1);
                    }
                    match &mut self.memory[reg.pointer] {
                        VMMemory::Number(num) => {
                            num.variable_id = operation.arg2 as u32;
                        },
                        VMMemory::String(str) => {
                            str.variable_id = operation.arg2 as u32;
                        },
                        VMMemory::Boolean(bool) => {
                            bool.variable_id = operation.arg2 as u32;
                        },
                        VMMemory::Null(null) => {
                            null.variable_id = operation.arg2 as u32;
                        },
                    }
                },
                29 => {
                    //RET
                    //not implemented yet
                },
                _ => panic!("Unknown opcode: {}", main_func.body[self.pc].opcode),
                
            }
        }
    }
}